package scanner

import error.Severity.ERROR
import org.junit.Assert.assertEquals
import org.junit.Assert.assertNotNull
import org.junit.Assert.assertTrue
import org.junit.Before
import org.junit.Test
import scanner.LispScanner.UnterminatedStringException
import testutil.TestUtilities.createInputStreamFromString
import token.AtSign
import token.Backquote
import token.Comma
import token.Eof
import token.Identifier
import token.LeftParenthesis
import token.Number
import token.QuoteMark
import token.QuotedString
import token.RightParenthesis
import token.Token
import token.TokenFactory.BadCharacterException

class LispScannerTypeTest {

    private var expectedTypes: MutableList<Class<out Token>>? = null

    private fun assertTokenTypesMatch(input: String) {
        val stringInputStream = createInputStreamFromString(input)
        val lispScanner = LispScanner(stringInputStream, "testFile")

        for (type in expectedTypes!!)
            assertTrue(type.isInstance(lispScanner.nextToken))

        assertTrue(lispScanner.nextToken is Eof)
    }

    @Before
    fun setUp() {
        expectedTypes = mutableListOf()
    }

    @Test
    fun givenEmptyFile_ReturnsCorrectTypes() {
        val input = ""

        assertTokenTypesMatch(input)
    }

    @Test(expected = BadCharacterException::class)
    fun givenBadCharacter_ThrowsException() {
        val input = "["

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenBadCharacter_ExceptionContainsCorrectSeverity() {
        val input = "abc\ndef["
        expectedTypes!!.add(Identifier::class.java)
        expectedTypes!!.add(Identifier::class.java)

        try {
            assertTokenTypesMatch(input)
        } catch (e: BadCharacterException) {
            assertEquals(ERROR, e.severity)
        }
    }

    @Test
    fun givenBadCharacter_ExceptionContainsMessage() {
        val input = "abc\ndef["
        expectedTypes!!.add(Identifier::class.java)
        expectedTypes!!.add(Identifier::class.java)

        try {
            assertTokenTypesMatch(input)
        } catch (e: BadCharacterException) {
            val message = e.message
            assertNotNull(message)
            assertTrue(message.length > 0)
        }
    }

    @Test
    fun givenNil_ReturnsCorrectTypes() {
        val input = "()"
        expectedTypes!!.add(LeftParenthesis::class.java)
        expectedTypes!!.add(RightParenthesis::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenListOfNumbers_ReturnsCorrectTypes() {
        val input = "(1 2)"
        expectedTypes!!.add(LeftParenthesis::class.java)
        expectedTypes!!.add(Number::class.java)
        expectedTypes!!.add(Number::class.java)
        expectedTypes!!.add(RightParenthesis::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenString_ReturnsCorrectTypes() {
        val input = "\"string\""
        expectedTypes!!.add(QuotedString::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenStringWithEscapedDoubleQuote_ReturnsCorrectTypes() {
        val input = "\"string \n hi \\\" bye\""
        expectedTypes!!.add(QuotedString::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenStringWithEscapedDoubleQuoteAndComment_ReturnsCorrectTypes() {
        val input = "\"string \n hi \\\" ; bye\""
        expectedTypes!!.add(QuotedString::class.java)

        assertTokenTypesMatch(input)
    }

    @Test(expected = UnterminatedStringException::class)
    fun givenUnterminatedString_ThrowsException() {
        val input = "\"oh no!"
        expectedTypes!!.add(QuotedString::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenUnterminatedString_ExceptionHasCorrectSeverity() {
        val input = "\"oh no!"
        expectedTypes!!.add(QuotedString::class.java)

        try {
            assertTokenTypesMatch(input)
        } catch (e: LispScanner.UnterminatedStringException) {
            assertEquals(ERROR, e.severity)
        }
    }

    @Test
    fun givenUnterminatedString_ExceptionContainsMessage() {
        val input = "\"oh no!"
        expectedTypes!!.add(QuotedString::class.java)

        try {
            assertTokenTypesMatch(input)
        } catch (e: LispScanner.UnterminatedStringException) {
            val message = e.message
            assertNotNull(message)
            assertTrue(message.length > 0)
        }
    }

    @Test
    fun givenIdentifier_ReturnsCorrectTypes() {
        val input = "abcdefgHIJKLMNOP1234"
        expectedTypes!!.add(Identifier::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenPrefixedIdentifiers_ReturnsCorrectTypes() {
        val input = "-a +b"
        expectedTypes!!.add(Identifier::class.java)
        expectedTypes!!.add(Identifier::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenSingleDigitNumber_ReturnsCorrectTypes() {
        val input = "1"
        expectedTypes!!.add(Number::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenPrefixedNegativeNumber_ReturnsCorrectTypes() {
        val input = "-1"
        expectedTypes!!.add(Number::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenPrefixedNumber_ReturnsCorrectTypes() {
        val input = "+1"
        expectedTypes!!.add(Number::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenPrefixedNegativeNumberAndIdentifier_ReturnsCorrectTypes() {
        val input = "-1apple"
        expectedTypes!!.add(Number::class.java)
        expectedTypes!!.add(Identifier::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenPrefixedNumberAndIdentifier_ReturnsCorrectTypes() {
        val input = "+1apple"
        expectedTypes!!.add(Number::class.java)
        expectedTypes!!.add(Identifier::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenPrefixedNegativeNumberAndString_ReturnsCorrectTypes() {
        val input = "-1\"apple\""
        expectedTypes!!.add(Number::class.java)
        expectedTypes!!.add(QuotedString::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenPrefixedNumberAndString_ReturnsCorrectTypes() {
        val input = "+1\"apple\""
        expectedTypes!!.add(Number::class.java)
        expectedTypes!!.add(QuotedString::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenMultipleDigitNumber_ReturnsCorrectTypes() {
        val input = "1234567890"
        expectedTypes!!.add(Number::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenQuote_ReturnsCorrectTypes() {
        val input = "'"
        expectedTypes!!.add(QuoteMark::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenManyTypesWithNoWhitespace_ReturnsCorrectTypes() {
        val input = "xxx\"hi\"999()'aaa"
        expectedTypes!!.add(Identifier::class.java)
        expectedTypes!!.add(QuotedString::class.java)
        expectedTypes!!.add(Number::class.java)
        expectedTypes!!.add(LeftParenthesis::class.java)
        expectedTypes!!.add(RightParenthesis::class.java)
        expectedTypes!!.add(QuoteMark::class.java)
        expectedTypes!!.add(Identifier::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenFunctionCall_ReturnsCorrectTypes() {
        val input = "(defun myFunction (x)\n  (print x))"
        expectedTypes!!.add(LeftParenthesis::class.java)
        expectedTypes!!.add(Identifier::class.java)
        expectedTypes!!.add(Identifier::class.java)
        expectedTypes!!.add(LeftParenthesis::class.java)
        expectedTypes!!.add(Identifier::class.java)
        expectedTypes!!.add(RightParenthesis::class.java)
        expectedTypes!!.add(LeftParenthesis::class.java)
        expectedTypes!!.add(Identifier::class.java)
        expectedTypes!!.add(Identifier::class.java)
        expectedTypes!!.add(RightParenthesis::class.java)
        expectedTypes!!.add(RightParenthesis::class.java)

        assertTokenTypesMatch(input)
    }

    @Test
    fun givenBackTickExpression_ReturnsCorrectTypes() {
        val input = "`(list ,a ,@b)"
        expectedTypes!!.add(Backquote::class.java)
        expectedTypes!!.add(LeftParenthesis::class.java)
        expectedTypes!!.add(Identifier::class.java)
        expectedTypes!!.add(Comma::class.java)
        expectedTypes!!.add(Identifier::class.java)
        expectedTypes!!.add(Comma::class.java)
        expectedTypes!!.add(AtSign::class.java)
        expectedTypes!!.add(Identifier::class.java)
        expectedTypes!!.add(RightParenthesis::class.java)

        assertTokenTypesMatch(input)
    }
}
